// (C) 2019-2022 GoodData Corporation
import { UnexpectedResponseError, } from "@gooddata/sdk-backend-spi";
import { isCatalogAttribute, isCatalogFact, isCatalogMeasure, isCatalogDateDataset, isCatalogAttributeHierarchy, } from "@gooddata/sdk-model";
import identity from "lodash/identity.js";
/**
 * @internal
 */
export class RecordedCatalogFactory {
    workspace;
    recordings;
    config;
    options;
    constructor(workspace, recordings = {}, config = {}, options = {
        types: ["attribute", "measure", "fact", "dateDataset", "attributeHierarchy"],
        excludeTags: [],
        includeTags: [],
        loadGroups: true,
    }) {
        this.workspace = workspace;
        this.recordings = recordings;
        this.config = config;
        this.options = options;
    }
    forDataset = (dataset) => {
        return this.withOptions({
            dataset,
        });
    };
    forTypes = (types) => {
        return this.withOptions({
            types,
        });
    };
    // include and exclude tags do not work yet
    includeTags = (tags) => {
        return this.withOptions({
            includeTags: tags,
        });
    };
    excludeTags = (tags) => {
        return this.withOptions({
            excludeTags: tags,
        });
    };
    withOptions = (options) => {
        const newOptions = {
            ...this.options,
            ...options,
        };
        return new RecordedCatalogFactory(this.workspace, this.recordings, this.config, newOptions);
    };
    withGroups(loadGroups) {
        return this.withOptions({
            loadGroups,
        });
    }
    load = async () => {
        const catalog = this.recordings.metadata?.catalog;
        if (!(catalog?.items && catalog?.groups)) {
            throw new UnexpectedResponseError("No catalog recording", 404, {});
        }
        const typeFilteredItems = catalog.items.filter((item) => this.options.types.includes(item.type));
        const catalogItems = typeFilteredItems.map(this.convertToCatalogItem.bind(this));
        return new RecordedCatalog(this.workspace, this.config, catalog.groups, catalogItems);
    };
    convertToCatalogItem(catalogItem) {
        if (isCatalogAttribute(catalogItem)) {
            return {
                ...catalogItem,
                attribute: {
                    ...catalogItem.attribute,
                    displayForms: catalogItem.displayForms,
                },
            };
        }
        return catalogItem;
    }
}
class RecordedCatalogBase {
    catalogGroups;
    items;
    constructor(catalogGroups, items) {
        this.catalogGroups = catalogGroups;
        this.items = items;
    }
    allItems() {
        return this.items;
    }
    attributes() {
        return this.items.filter(isCatalogAttribute);
    }
    measures() {
        return this.items.filter(isCatalogMeasure);
    }
    facts() {
        return this.items.filter(isCatalogFact);
    }
    dateDatasets() {
        return this.items.filter(isCatalogDateDataset);
    }
    groups() {
        return this.catalogGroups;
    }
    attributeHierarchies() {
        return this.items.filter(isCatalogAttributeHierarchy);
    }
}
class RecordedCatalog extends RecordedCatalogBase {
    workspace;
    config;
    constructor(workspace, config, catalogGroups, items) {
        super(catalogGroups, items);
        this.workspace = workspace;
        this.config = config;
    }
    availableItems() {
        return new RecordedAvailableCatalogFactory(this.workspace, this.config, this.catalogGroups, this.items);
    }
}
class RecordedAvailableCatalogFactory {
    workspace;
    config;
    groups;
    items;
    options;
    constructor(workspace, config, groups, items, options = {
        types: ["attribute", "measure", "fact", "dateDataset", "attributeHierarchy"],
        excludeTags: [],
        includeTags: [],
        loadGroups: true,
    }) {
        this.workspace = workspace;
        this.config = config;
        this.groups = groups;
        this.items = items;
        this.options = options;
    }
    // eslint-disable-next-line sonarjs/no-identical-functions
    forDataset = (dataset) => {
        return this.withOptions({
            dataset,
        });
    };
    // eslint-disable-next-line sonarjs/no-identical-functions
    forTypes = (types) => {
        return this.withOptions({
            types,
        });
    };
    // include and exclude tags do not work yet
    // eslint-disable-next-line sonarjs/no-identical-functions
    includeTags = (tags) => {
        return this.withOptions({
            includeTags: tags,
        });
    };
    // eslint-disable-next-line sonarjs/no-identical-functions
    excludeTags = (tags) => {
        return this.withOptions({
            excludeTags: tags,
        });
    };
    withOptions = (options) => {
        const newOptions = {
            ...this.options,
            ...options,
        };
        return new RecordedAvailableCatalogFactory(this.workspace, this.config, this.groups, this.items, newOptions);
    };
    forItems = (_items) => {
        // availability not implemented yet
        return this;
    };
    forInsight = (_insight) => {
        // availability not implemented yet
        return this;
    };
    // eslint-disable-next-line sonarjs/no-identical-functions
    withGroups(loadGroups) {
        return this.withOptions({
            loadGroups,
        });
    }
    load = async () => {
        return new RecordedAvailableCatalog(this.workspace, this.config, this.groups, this.items);
    };
}
class RecordedAvailableCatalog extends RecordedCatalogBase {
    workspace;
    config;
    filteredAttributes;
    filteredMeasures;
    filteredFacts;
    filteredDateDatasets;
    filteredAttributeHierarchies;
    constructor(workspace, config, groups, items) {
        super(groups, items);
        this.workspace = workspace;
        this.config = config;
        // without this inference starts thinking the identity may return undefined :/
        const typedIdentity = identity;
        const { availableAttributes = typedIdentity, availableMeasures = typedIdentity, availableFacts = typedIdentity, availableDateDatasets = typedIdentity, availableAttributeHierarchies = typedIdentity, } = this.config.catalogAvailability ?? {};
        this.filteredAttributes = availableAttributes(this.attributes());
        this.filteredMeasures = availableMeasures(this.measures());
        this.filteredFacts = availableFacts(this.facts());
        this.filteredDateDatasets = availableDateDatasets(this.dateDatasets());
        this.filteredAttributeHierarchies = availableAttributeHierarchies(this.attributeHierarchies());
    }
    allAvailableItems = () => {
        return [
            ...this.filteredAttributes,
            ...this.filteredMeasures,
            ...this.filteredFacts,
            ...this.filteredDateDatasets,
            ...this.filteredAttributeHierarchies,
        ];
    };
    availableAttributes = () => {
        return this.filteredAttributes;
    };
    availableMeasures = () => {
        return this.filteredMeasures;
    };
    availableFacts() {
        return this.filteredFacts;
    }
    availableDateDatasets() {
        return this.filteredDateDatasets;
    }
    availableAttributeHierarchies() {
        return this.filteredAttributeHierarchies;
    }
}
//# sourceMappingURL=catalog.js.map